/*
 * Brinzea Razvan-Stefan
 * Grupa 333CB
 * Facultatea de Automatica si Calculatoare
 * Proiect PM - profesorul de pian
 */

#include "lcd.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <avr/io.h>
#include <util/delay.h>

#include <avr/interrupt.h>

#define DEBOUNCE_TIME 150
#define BASE_DURATION 250
#define MIN_DURATION 100
#define MAX_DURATION 350
#define BASE_PAUSE 20
#define MIN_PAUSE 10
#define MAX_PAUSE 260
#define BAR_SIZE 4

enum state_t { INTRO, SONG_SELECT, PAUSE_SELECT, SECTION_SELECT, OCTAVE_SELECT, DURATION_SELECT };
int no_states = 6;

typedef struct song {
	char name[16];
	char notes[128];
	int *frequencies;
	int duration;
	int quarter_notes;
} song_t;

/* Intreruperi pentru difuzor (pinul PWM PB3) */
ISR(TIMER0_OVF_vect)
{
    PORTB |= (1 << PB3);
}

ISR(TIMER0_COMPB_vect)
{
    PORTB &= ~(1 << PB3);
}

/* Variabile globale ce definesc starea aplicatiei. */
enum state_t current_state = INTRO;
song_t song1, song2, song3, song4;
song_t *songs[] = { &song1, &song2, &song3, &song4 };
int no_songs = 4;
song_t *current_song = songs[0];
int current_song_index = 0;

int lower_frequencies[] = { 131, 147, 165, 175, 196, 220, 247, 261 };
int base_frequencies[] = { 261, 293, 329, 349, 392, 440, 494, 523 };
int higher_frequencies[] = { 523, 587, 659, 698, 784, 880, 988, 1047 } ;
int current_duration = BASE_DURATION;
int current_note = -1;
int current_pause = BASE_PAUSE;
int *current_frequencies = base_frequencies;
int current_section = -1;

/* Initializeaza o structura de tip song_t care are memoria deja alocata. */
void init_song(song_t *song, char *name, char *notes, int *frequencies, int duration, int quarter_notes) {
	strcpy(song->name, name);
	strcpy(song->notes, notes);
	song->frequencies = frequencies;
	song->duration = duration;
	song->quarter_notes = quarter_notes;
}

/* Afiseaza un mesaj pe linia de sus a LCD-ului. */
void LCD_clearAndPrintTop(const char *msg)
{
	LCD_printAt(0, "                ");
	LCD_printAt(0, msg);
}

/* Afiseaza un mesaj pe linia de jos a LCD-ului. */
void LCD_clearAndPrintBottom(const char *msg)
{
	LCD_printAt(LCD_INSTR_nextLine, "                ");
	LCD_printAt(LCD_INSTR_nextLine, msg);
}

/* In functie de starea curenta, afiseaza informatiile relevante pe ecran. */
void update_screen() {
	switch(current_state) {
	case INTRO:
		LCD_clearAndPrintTop("Brinzea Razvan");
		LCD_clearAndPrintBottom("Grupa 333CB");
		break;
	
	case SONG_SELECT:
		LCD_clearAndPrintTop("Select song");
		char song_string[16];
		
		sprintf(song_string, ">%d: %s", current_song_index, current_song->name);
		
		LCD_clearAndPrintBottom(song_string);
		break;
		
	case PAUSE_SELECT:
		LCD_clearAndPrintTop("Change pause");
		char pause_string[16];
		
		sprintf(pause_string, "> %d ms", current_pause);
		
		LCD_clearAndPrintBottom(pause_string);
		break;
		
	case SECTION_SELECT:
		LCD_clearAndPrintTop("Select section");
		char section_string[16];
		
		if (current_section == -1)
			sprintf(section_string, "> Full song");
		else
			sprintf(section_string, "> section %d", current_section);
		
		LCD_clearAndPrintBottom(section_string);
		break;
		
	case OCTAVE_SELECT:
		LCD_clearAndPrintTop("Change octave");
		char octave_string[16];
		
		if (current_frequencies == lower_frequencies)
			sprintf(octave_string, "> Low");
		else if (current_frequencies == base_frequencies)
			sprintf(octave_string, "> Medium");
		else if (current_frequencies == higher_frequencies)
			sprintf(octave_string, "> High");
		else
			sprintf(octave_string, "> Custom");
		
		LCD_clearAndPrintBottom(octave_string);
		break;
		
	case DURATION_SELECT:
		LCD_clearAndPrintTop("Change duration");
		char duration_string[16];
		sprintf(duration_string, "> %d", current_duration);
		
		LCD_clearAndPrintBottom(duration_string);
		break;
	}
}

/* Trece la urmatoarea stare si actualizeaza ecranul. */
void next_state() {
	switch(current_state) {
	case INTRO:
		current_state = SONG_SELECT;
		break;
	
	case SONG_SELECT:
		current_state = SECTION_SELECT;
		break;
	
	case SECTION_SELECT:
		current_state = PAUSE_SELECT;
		break;
	
	case PAUSE_SELECT:
		current_state = OCTAVE_SELECT;
		break;
		
	case OCTAVE_SELECT:
		current_state = DURATION_SELECT;
		break;
		
	case DURATION_SELECT:
		current_state = INTRO;
		break;
	}
	
	update_screen();
}

/* Selecteaza un nou cantec din lista. */
void select_song(int index) {
	current_song = songs[index];
	
	current_frequencies = current_song->frequencies;
	current_duration = current_song->duration;
	current_section = -1;
}

/* Trece la valoarea urmatoare in cadrul meniului curent. */
void next_value() {
	switch(current_state) {
	case INTRO:
		break;
	
	case SONG_SELECT:
		current_song_index = (current_song_index + 1) % no_songs;
		select_song(current_song_index);
	
		break;
		
	case PAUSE_SELECT:
		if (current_pause < MAX_PAUSE)
			current_pause += 50;
		break;
		
	case SECTION_SELECT:
		if (current_section == (current_song->quarter_notes / BAR_SIZE) - 1)
			current_section = -1;
		else current_section++;
		break;
		
	case OCTAVE_SELECT:
		if (current_frequencies == lower_frequencies)
			current_frequencies = base_frequencies;
		else if (current_frequencies == base_frequencies)
			current_frequencies = higher_frequencies;
		break;
		
	case DURATION_SELECT:
		if (current_duration < MAX_DURATION)
			current_duration += 50;
		break;
	}
	
	update_screen();
}

/* Trece la valoarea anterioara in cadrul meniului curent. */
void prev_value() {
	switch(current_state) {
	case INTRO:
		break;
	
	case SONG_SELECT:
		if (current_song_index == 0)
			current_song_index = no_songs - 1;
		else
			current_song_index--;
		
		select_song(current_song_index);
		
		break;
		
	case PAUSE_SELECT:
		if (current_pause > MIN_PAUSE)
			current_pause -= 50;
		break;
		
	case SECTION_SELECT:
		if (current_section == -1)
			current_section = current_song->quarter_notes / BAR_SIZE - 1;
		else
			current_section--;
		break;
		
	case OCTAVE_SELECT:
		if (current_frequencies == higher_frequencies)
			current_frequencies = base_frequencies;
		else if (current_frequencies == base_frequencies)
			current_frequencies = lower_frequencies;
		break;
		
	case DURATION_SELECT:
		if (current_duration > MIN_DURATION)
			current_duration -= 50;
		break;
	}
	
	update_screen();
}

/* Actualizeaza registrii timer-ului 0 pentru a schimba frecventa notei redate. */
void update_song(int frequency)
{
	uint16_t new_top = 16000000 / 1024 / 2 / frequency;
	TCNT0 = 0;
	OCR0B = (uint8_t)new_top * OCR0B / OCR0A;
	OCR0A = new_top;
}

/* Reda un anumit sunet pentru o anumita perioada de timp. */
void play_sound(int duration, int frequency) {
	
	/* Porneste timer-ul 0. */
	TCCR0A |= (1 << COM0A0) | (1 << COM0B0) | (1 << WGM01);
	
	/* Actualizeaza registrii timer-ului. */
	update_song(frequency);
	
	/* Asteapta pentru durata specificata. */
	_delay_ms(duration);
	
	/* Opreste timer-ul. */
	TCCR0A = 0;
}

/* Aprinde unul dintre cele 7 LED-uri de pe claviatura. */
/* LED-urile sunt legate sub forma unei matrice. */
void light_led(int index) {
	
	switch(index) {
	case -1:
		/* Opreste toate LED-urile. */
		PORTD &= ~(1 << PD0);
		PORTD &= ~(1 << PD4);
		PORTD &= ~(1 << PD1);
		PORTD &= ~(1 << PD5);
		PORTD &= ~(1 << PD6);
		PORTD &= ~(1 << PD7);
		break;
		
	case 0:
		/* PD0 is ON, PD4 is OFF. */
		PORTD |= (1 << PD0);
		PORTD &= ~(1 << PD4);
		
		PORTD &= ~(1 << PD1);
		PORTD |= (1 << PD5);
		PORTD |= (1 << PD6);
		PORTD |= (1 << PD7);
		break;
		
	case 1:
		/* PD1 is ON, PD4 is OFF. */
		PORTD |= (1 << PD1);
		PORTD &= ~(1 << PD4);
		
		PORTD &= ~(1 << PD0);
		PORTD |= (1 << PD5);
		PORTD |= (1 << PD6);
		PORTD |= (1 << PD7);
		break;
		
	case 2:
		/* PD0 is ON, PD5 is OFF. */
		PORTD |= (1 << PD0);
		PORTD &= ~(1 << PD5);
		
		PORTD &= ~(1 << PD1);
		PORTD |= (1 << PD4);
		PORTD |= (1 << PD6);
		PORTD |= (1 << PD7);
		break;
		
	case 3:
		/* PD1 is ON, PD5 is OFF. */
		PORTD |= (1 << PD1);
		PORTD &= ~(1 << PD5);
		
		PORTD &= ~(1 << PD0);
		PORTD |= (1 << PD4);
		PORTD |= (1 << PD6);
		PORTD |= (1 << PD7);
		break;
		
	case 4:
		/* PD0 is ON, PD6 is OFF. */
		PORTD |= (1 << PD0);
		PORTD &= ~(1 << PD6);
		
		PORTD &= ~(1 << PD1);
		PORTD |= (1 << PD5);
		PORTD |= (1 << PD4);
		PORTD |= (1 << PD7);
		break;
		
	case 5:
		/* PD1 is ON, PD6 is OFF. */
		PORTD |= (1 << PD1);
		PORTD &= ~(1 << PD6);
		
		PORTD &= ~(1 << PD0);
		PORTD |= (1 << PD5);
		PORTD |= (1 << PD4);
		PORTD |= (1 << PD7);
		break;
		
	case 6:
		/* PD0 is ON, PD7 is OFF. */
		PORTD |= (1 << PD0);
		PORTD &= ~(1 << PD7);
		
		PORTD &= ~(1 << PD1);
		PORTD |= (1 << PD5);
		PORTD |= (1 << PD4);
		PORTD |= (1 << PD6);
		break;
		
	case 7:
		/* PD1 is ON, PD7 is OFF. */
		PORTD |= (1 << PD1);
		PORTD &= ~(1 << PD7);
		
		PORTD &= ~(1 << PD0);
		PORTD |= (1 << PD5);
		PORTD |= (1 << PD4);
		PORTD |= (1 << PD6);
		break;
		
	default:
		break;
	}
}

/* Reda o anumita nota pentru o durata de timp data. */
/* In acelasi timp, se aprinde si LED-ul corespunzator tastei de pe claviatura. */
void play_note(int duration, int frequencies[], int index) {
	if (index < 0 || index > 7)
		return;
	
	light_led(index);
	
	play_sound(duration, frequencies[index]);
	
	light_led(-1);
}

/* Incepe redarea continua a unei note si aprinde LED-ul corespunzator. */
void start_note(int frequencies[], int index) {
	if (index < 0 || index > 7)
		return;
	
	light_led(index);
	
	/* Porneste timer-ul. */
	TCCR0A |= (1 << COM0A0) | (1 << COM0B0) | (1 << WGM01);
	
	/* Actualizeaza registrii timer-ului. */
	update_song(frequencies[index]);
	
	current_note = index;
}

/* Opreste redarea oricarei note si stinge LED-urile. */
void end_note() {
	TCCR0A = 0;
	light_led(-1);
	current_note = -1;
}

/* Reda in intregime melodia selectata. */
void play_current_song() {
	int i;
	int start_index = 0;
	int end_index = strlen(current_song->notes);
	
	/* Daca se doreste redarea unei secvente, se va selecta un subsir. */
	if (current_section != -1) {
		char *it = current_song->notes;
		
		for (i = 0; i <= current_section; i++) {
			int half_notes = 0;
			
			start_index = (it - current_song->notes);
			while (half_notes < 2 * BAR_SIZE) {
				if (*(it + 1) == 'f')
					half_notes += 2;
				else
					half_notes += 1;
				
				it += 2;
			}
			end_index = (it - current_song->notes);
		}
	}
	
	/* Se va itera prin sirul ales, si se va reda un sunet pentru fiecare nota. */
	for (i = start_index; i < end_index; i+= 2) {
		if (current_song->notes[i] >= '1' && current_song->notes[i] <= '8') {
			if (current_song->notes[i+1] == 'f')
				play_note(current_duration, current_frequencies, current_song->notes[i] - '1');
			else if (current_song->notes[i+1] == 'h')
				play_note(current_duration / 2, current_frequencies, current_song->notes[i] - '1');
		}
		else {
			if (current_song->notes[i+1] == 'f') {
				_delay_ms(current_duration);
			}
			else if (current_song->notes[i+1] == 'h') {
				_delay_ms(current_duration / 2);
			}
		}
		
		_delay_ms(current_pause);
	}
}

/* Intreruperea pentru butoanele de control */
ISR(PCINT1_vect)
{
    // PB0 apasat
    if((PINB & (1 << PB0)) == 0)
    {
        prev_value();
		_delay_ms(DEBOUNCE_TIME);
    }
    
    // PB1 apasat
    if((PINB & (1 << PB1)) == 0)
    {
		next_state();
		_delay_ms(DEBOUNCE_TIME);
    }
	
	// PB2 apasat
    if((PINB & (1 << PB2)) == 0)
    {
        play_current_song();
		_delay_ms(DEBOUNCE_TIME);
    }
	
	// PB3 apasat
    if((PINB & (1 << PB4)) == 0)
    {
        next_value();
		_delay_ms(DEBOUNCE_TIME);
    }
}

/* Initializeaza intrarile si iesirile */
void IO_init()
{
	/* Butoanele de control sunt iesiri (cu rezistente pull-up). */
	/* PB0, PB1, PB2, PB4 */
    DDRB &= ~((1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB4));
    PORTB |= ((1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB4));
    
	/* Butoanele pentru claviatura sunt iesiri (cu rezistente pull-up). */
	/* PA0, PA1, PA2, PA3, PA4, PC0, PC1, PC2 */
	DDRA &= ~((1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4));
	PORTA |= ((1 << PA0) | (1 << PA1) | (1 << PA2) | (1 << PA3) | (1 << PA4));
	DDRC &= ~((1 << PC0) | (1 << PC1) | (1 << PC2));
	PORTC |= ((1 << PC0) | (1 << PC1) | (1 << PC2));
	
	/* Difuzorul este iesire: PB3 */
	DDRB |= (1 << PB3);
	
	/* LED-urile pentru claviatura sunt iesiri: PD0, PD1, PD4, PD5, PD6, PD7 */
	DDRD |= (1 << PD0) | (1 << PD1) | (1 << PD4) | (1 << PD5) | (1 << PD6) | (1 << PD7);
	
    /* Intreruperile sunt activate pentru: */
	/* Butoanele de control: PB0, PB1, PB2, PB4 */
    PCICR  = (1 << PCIE1);
    PCMSK1 = (1 << PCINT8) | (1 << PCINT9) | (1 << PCINT10) | (1 << PCINT12);
}

/* Timer pentru difuzor (PWM) */
void pwm_init()
{
    /* Timer0 este configurat cu modul CTC, top OCR0A si prescaler 1024 */
    TCCR0A |= (1 << COM0A0) | (1 << COM0B0) | (1 << WGM01);
    TCCR0B |= (5 << CS00);

    OCR0B = 0;
    OCR0A = 0;
	
	/* Dezactiveaza sunetul pentru moment. */
	TCCR0A = 0;
}

int main()
{
	/* Activeaza intreruperile. */
	sei();
	
	/* Initializeaza intrarile si iesirile. */
	IO_init();
	
	/* Initializeaza ecranul LCD. */
	LCD_init();

	/* Afiseaza un mesaj initial. */
	LCD_clearAndPrintTop("Brinzea Razvan");
	LCD_clearAndPrintBottom("Grupa 333CB");
	
	/* initializeaza cantecele preincarcate */
	init_song(&song1, "Twinkle Twinkle", "1f1f5f5f6f6f5f*f4f4f3f3f2f2f1f", base_frequencies, 400, 16);
	init_song(&song2, "Jingle Bells", "3f3f3f*f3f3f3f*f3f5f1f2f3f*f*f*f", base_frequencies, 250, 16);
	init_song(&song3, "Godfather", "*f3f6f8f7f6f8f6f7f6f4f5f3f*f*f*f", base_frequencies, 250, 16);
	init_song(&song4, "Iron Man", "3f*f5f*f5f6f6f*f8h7h8h7h8h7h5f5f6f6f*f", lower_frequencies, 350, 16);
	select_song(current_song_index);
	
	/* Initializeaza registrii pentru timer-ul 0. */
	pwm_init();
	
	/* Sistem de polling pentru claviatura. */
	while(1) {
			// PA0 apasat
			if ((PINA & (1 << PA0)) == 0)
			{
				if (current_note != 0)
					start_note(current_frequencies, 0);
			}
			
			// PA1 apasat
			else if ((PINA & (1 << PA1)) == 0)
			{
				if (current_note != 1)
					start_note(current_frequencies, 1);
			}
			
			// PA2 apasat
			else if ((PINA & (1 << PA2)) == 0)
			{
				if (current_note != 2)
					start_note(current_frequencies, 2);
			}
			
			// PA3 apasat
			else if ((PINA & (1 << PA3)) == 0)
			{
				if (current_note != 3)
					start_note(current_frequencies, 3);
			}
			
			// PA4 apasat
			else if ((PINA & (1 << PA4)) == 0)
			{
				if (current_note != 4)
					start_note(current_frequencies, 4);
			}
			
			// PC0 apasat
			else if ((PINC & (1 << PC0)) == 0)
			{
				if (current_note != 5)
					start_note(current_frequencies, 5);
			}
			
			// PC1 apasat
			else if ((PINC & (1 << PC1)) == 0)
			{
				if (current_note != 6)
					start_note(current_frequencies, 6);
			}
			
			// PC2 apasat
			else if ((PINC & (1 << PC2)) == 0)
			{
				if (current_note != 7)
					start_note(current_frequencies, 7);
			}
			
			else {
				end_note();
			}
			
			_delay_ms(1);
	}
	
    return 0;
}

